// Generated from player/lua/console.lua

"-- Copyright (C) 2019 the mpv developers\n"
"--\n"
"-- Permission to use, copy, modify, and/or distribute this software for any\n"
"-- purpose with or without fee is hereby granted, provided that the above\n"
"-- copyright notice and this permission notice appear in all copies.\n"
"--\n"
"-- THE SOFTWARE IS PROVIDED \"AS IS\" AND THE AUTHOR DISCLAIMS ALL WARRANTIES\n"
"-- WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF\n"
"-- MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY\n"
"-- SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES\n"
"-- WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION\n"
"-- OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN\n"
"-- CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.\n"
"\n"
"local utils = require 'mp.utils'\n"
"local assdraw = require 'mp.assdraw'\n"
"\n"
"-- Default options\n"
"local opts = {\n"
"    font = \"\",\n"
"    font_size = 16,\n"
"    border_size = 1,\n"
"    case_sensitive = true,\n"
"    history_dedup = true,\n"
"    font_hw_ratio = 'auto',\n"
"}\n"
"\n"
"local function detect_platform()\n"
"    local platform = mp.get_property_native('platform')\n"
"    if platform == 'darwin' or platform == 'windows' then\n"
"        return platform\n"
"    elseif os.getenv('WAYLAND_DISPLAY') then\n"
"        return 'wayland'\n"
"    end\n"
"    return 'x11'\n"
"end\n"
"\n"
"-- Pick a better default font for Windows and macOS\n"
"local platform = detect_platform()\n"
"if platform == 'windows' then\n"
"    opts.font = 'Consolas'\n"
"    opts.case_sensitive = false\n"
"elseif platform == 'darwin' then\n"
"    opts.font = 'Menlo'\n"
"else\n"
"    opts.font = 'monospace'\n"
"end\n"
"\n"
"-- Apply user-set options\n"
"require 'mp.options'.read_options(opts)\n"
"\n"
"local styles = {\n"
"    -- Colors are stolen from base16 Eighties by Chris Kempson\n"
"    -- and converted to BGR as is required by ASS.\n"
"    -- 2d2d2d 393939 515151 697374\n"
"    -- 939fa0 c8d0d3 dfe6e8 ecf0f2\n"
"    -- 7a77f2 5791f9 66ccff 99cc99\n"
"    -- cccc66 cc9966 cc99cc 537bd2\n"
"\n"
"    debug = '{\\\\1c&Ha09f93&}',\n"
"    v = '{\\\\1c&H99cc99&}',\n"
"    warn = '{\\\\1c&H66ccff&}',\n"
"    error = '{\\\\1c&H7a77f2&}',\n"
"    fatal = '{\\\\1c&H5791f9&}',\n"
"    suggestion = '{\\\\1c&Hcc99cc&}',\n"
"    selected_suggestion = '{\\\\1c&H2fbdfa&\\\\b1}',\n"
"    disabled = '{\\\\1c&Hcccccc&}',\n"
"}\n"
"\n"
"local terminal_styles = {\n"
"    debug = '\\027[90m',\n"
"    v = '\\027[32m',\n"
"    warn = '\\027[33m',\n"
"    error = '\\027[31m',\n"
"    fatal = '\\027[91m',\n"
"    selected_suggestion = '\\027[7m',\n"
"    disabled = '\\027[38;5;8m',\n"
"}\n"
"\n"
"local repl_active = false\n"
"local insert_mode = false\n"
"local pending_update = false\n"
"local line = ''\n"
"local cursor = 1\n"
"local default_prompt = '>'\n"
"local prompt = default_prompt\n"
"local default_id = 'default'\n"
"local id = default_id\n"
"local histories = {[id] = {}}\n"
"local history = histories[id]\n"
"local history_pos = 1\n"
"local searching_history = false\n"
"local log_buffers = {[id] = {}}\n"
"local key_bindings = {}\n"
"local global_margins = { t = 0, b = 0 }\n"
"local input_caller\n"
"\n"
"local suggestion_buffer = {}\n"
"local selected_suggestion_index\n"
"local completion_pos\n"
"local completion_append\n"
"local path_separator = platform == 'windows' and '\\\\' or '/'\n"
"local completion_old_line\n"
"local completion_old_cursor\n"
"\n"
"local selectable_items\n"
"local matches = {}\n"
"local selected_match = 1\n"
"local first_match_to_print = 1\n"
"\n"
"local set_active\n"
"\n"
"\n"
"-- Naive helper function to find the next UTF-8 character in 'str' after 'pos'\n"
"-- by skipping continuation bytes. Assumes 'str' contains valid UTF-8.\n"
"local function next_utf8(str, pos)\n"
"    if pos > str:len() then return pos end\n"
"    repeat\n"
"        pos = pos + 1\n"
"    until pos > str:len() or str:byte(pos) < 0x80 or str:byte(pos) > 0xbf\n"
"    return pos\n"
"end\n"
"\n"
"-- As above, but finds the previous UTF-8 character in 'str' before 'pos'\n"
"local function prev_utf8(str, pos)\n"
"    if pos <= 1 then return pos end\n"
"    repeat\n"
"        pos = pos - 1\n"
"    until pos <= 1 or str:byte(pos) < 0x80 or str:byte(pos) > 0xbf\n"
"    return pos\n"
"end\n"
"\n"
"local function len_utf8(str)\n"
"    local len = 0\n"
"    local pos = 1\n"
"    while pos <= str:len() do\n"
"        pos = next_utf8(str, pos)\n"
"        len = len + 1\n"
"    end\n"
"    return len\n"
"end\n"
"\n"
"local function truncate_utf8(str, max_length)\n"
"    local len = 0\n"
"    local pos = 1\n"
"    while pos <= #str do\n"
"        local last_pos = pos\n"
"        pos = next_utf8(str, pos)\n"
"        len = len + 1\n"
"        if pos > last_pos + 1 then\n"
"            if len == max_length - 1 then\n"
"                pos = last_pos\n"
"            else\n"
"                len = len + 1\n"
"            end\n"
"        end\n"
"        if len == max_length - 1 then\n"
"            return str:sub(1, pos - 1) .. '\342\213\257'\n"
"        end\n"
"    end\n"
"    return str\n"
"end\n"
"\n"
"\n"
"-- Functions to calculate the font width.\n"
"local width_length_ratio = 0.5\n"
"local osd_width, osd_height = 100, 100\n"
"\n"
"---Update osd resolution if valid\n"
"local function update_osd_resolution()\n"
"    local dim = mp.get_property_native('osd-dimensions')\n"
"    if not dim or dim.w == 0 or dim.h == 0 then\n"
"        return\n"
"    end\n"
"    osd_width = dim.w\n"
"    osd_height = dim.h\n"
"end\n"
"\n"
"local text_osd = mp.create_osd_overlay('ass-events')\n"
"text_osd.compute_bounds, text_osd.hidden = true, true\n"
"\n"
"local function measure_bounds(ass_text)\n"
"    update_osd_resolution()\n"
"    text_osd.res_x, text_osd.res_y = osd_width, osd_height\n"
"    text_osd.data = ass_text\n"
"    local res = text_osd:update()\n"
"    return res.x0, res.y0, res.x1, res.y1\n"
"end\n"
"\n"
"---Measure text width and normalize to a font size of 1\n"
"---text has to be ass safe\n"
"local function normalized_text_width(text, size, horizontal)\n"
"    local align, rotation = horizontal and 7 or 1, horizontal and 0 or -90\n"
"    local template = '{\\\\pos(0,0)\\\\rDefault\\\\blur0\\\\bord0\\\\shad0\\\\q2\\\\an%s\\\\fs%s\\\\fn%s\\\\frz%s}%s'\n"
"    size = size / 0.8\n"
"    local width\n"
"    -- Limit to 5 iterations\n"
"    local repetitions_left = 5\n"
"    for i = 1, repetitions_left do\n"
"        size = size * 0.8\n"
"        local ass = assdraw.ass_new()\n"
"        ass.text = template:format(align, size, opts.font, rotation, text)\n"
"        local _, _, x1, y1 = measure_bounds(ass.text)\n"
"        -- Check if nothing got clipped\n"
"        if x1 and x1 < osd_width and y1 < osd_height then\n"
"            width = horizontal and x1 or y1\n"
"            break\n"
"        end\n"
"        if i == repetitions_left then\n"
"            width = 0\n"
"        end\n"
"    end\n"
"    return width / size, horizontal and osd_width or osd_height\n"
"end\n"
"\n"
"local function fit_on_osd(text)\n"
"    local estimated_width = #text * width_length_ratio\n"
"    if osd_width >= osd_height then\n"
"        -- Fill the osd as much as possible, bigger is more accurate.\n"
"        return math.min(osd_width / estimated_width, osd_height), true\n"
"    else\n"
"        return math.min(osd_height / estimated_width, osd_width), false\n"
"    end\n"
"end\n"
"\n"
"local measured_font_hw_ratio = nil\n"
"local function get_font_hw_ratio()\n"
"    local font_hw_ratio = tonumber(opts.font_hw_ratio)\n"
"    if font_hw_ratio then\n"
"        return font_hw_ratio\n"
"    end\n"
"    if not measured_font_hw_ratio then\n"
"        local alphabet = 'abcdefghijklmnopqrstuvwxyz'\n"
"        local text = alphabet:rep(3)\n"
"        update_osd_resolution()\n"
"        local size, horizontal = fit_on_osd(text)\n"
"        local normalized_width = normalized_text_width(text, size * 0.9, horizontal)\n"
"        measured_font_hw_ratio = #text / normalized_width * 0.95\n"
"    end\n"
"    return measured_font_hw_ratio\n"
"end\n"
"\n"
"\n"
"-- Escape a string for verbatim display on the OSD\n"
"local function ass_escape(str)\n"
"    return mp.command_native({'escape-ass', str})\n"
"end\n"
"\n"
"local function calculate_max_log_lines()\n"
"    if not mp.get_property_native('vo-configured')\n"
"       or not mp.get_property_native('video-osd') then\n"
"        -- Subtract 1 for the input line and for each line in the status line.\n"
"        -- This does not detect wrapped lines.\n"
"        return mp.get_property_native('term-size/h', 24) - 2 -\n"
"               select(2, mp.get_property('term-status-msg'):gsub('\\\\n', ''))\n"
"    end\n"
"\n"
"    return math.floor(mp.get_property_native('osd-height')\n"
"                      / mp.get_property_native('display-hidpi-scale', 1)\n"
"                      * (1 - global_margins.t - global_margins.b)\n"
"                      / opts.font_size\n"
"                      -- Subtract 1 for the input line and 1 for the newline\n"
"                      -- between the log and the input line.\n"
"                      - 2)\n"
"end\n"
"\n"
"-- Takes a list of strings, a max width in characters and\n"
"-- optionally a max row count.\n"
"-- The result contains at least one column.\n"
"-- Rows are cut off from the top if rows_max is specified.\n"
"-- returns a string containing the formatted table and the row count\n"
"local function format_table(list, width_max, rows_max)\n"
"    if #list == 0 then\n"
"        return '', 0\n"
"    end\n"
"\n"
"    local spaces_min = 2\n"
"    local spaces_max = 8\n"
"    local list_size = #list\n"
"    local column_count = 1\n"
"    local row_count = list_size\n"
"    local column_widths\n"
"    -- total width without spacing\n"
"    local width_total = 0\n"
"\n"
"    local list_widths = {}\n"
"    for i, item in ipairs(list) do\n"
"        list_widths[i] = len_utf8(item)\n"
"    end\n"
"\n"
"    -- use as many columns as possible\n"
"    for columns = 2, list_size do\n"
"        local rows_lower_bound = math.min(rows_max, math.ceil(list_size / columns))\n"
"        local rows_upper_bound = math.min(rows_max, list_size,\n"
"                                          math.ceil(list_size / (columns - 1) - 1))\n"
"        for rows = rows_upper_bound, rows_lower_bound, -1 do\n"
"            local cw = {}\n"
"            width_total = 0\n"
"\n"
"            -- find out width of each column\n"
"            for column = 1, columns do\n"
"                local width = 0\n"
"                for row = 1, rows do\n"
"                    local i = row + (column - 1) * rows\n"
"                    local item_width = list_widths[i]\n"
"                    if not item_width then break end\n"
"                    if width < item_width then\n"
"                        width = item_width\n"
"                    end\n"
"                end\n"
"                cw[column] = width\n"
"                width_total = width_total + width\n"
"                if width_total + (columns - 1) * spaces_min > width_max then\n"
"                    break\n"
"                end\n"
"            end\n"
"\n"
"            if width_total + (columns - 1) * spaces_min <= width_max then\n"
"                row_count = rows\n"
"                column_count = columns\n"
"                column_widths = cw\n"
"            else\n"
"                break\n"
"            end\n"
"        end\n"
"        if width_total + (columns - 1) * spaces_min > width_max then\n"
"            break\n"
"        end\n"
"    end\n"
"\n"
"    local spaces = math.floor((width_max - width_total) / (column_count - 1))\n"
"    spaces = math.max(spaces_min, math.min(spaces_max, spaces))\n"
"    local spacing = column_count > 1\n"
"                    and ass_escape(string.format('%' .. spaces .. 's', ' '))\n"
"                    or ''\n"
"\n"
"    local rows = {}\n"
"    for row = 1, row_count do\n"
"        local columns = {}\n"
"        for column = 1, column_count do\n"
"            local i = row + (column - 1) * row_count\n"
"            if i > #list then break end\n"
"            -- more then 99 leads to 'invalid format (width or precision too long)'\n"
"            local format_string = column == column_count and '%s'\n"
"                                  or '%-' .. math.min(column_widths[column], 99) .. 's'\n"
"            columns[column] = ass_escape(string.format(format_string, list[i]))\n"
"\n"
"            if i == selected_suggestion_index then\n"
"                columns[column] = styles.selected_suggestion .. columns[column]\n"
"                                  .. '{\\\\b0}'.. styles.suggestion\n"
"            end\n"
"        end\n"
"        -- first row is at the bottom\n"
"        rows[row_count - row + 1] = table.concat(columns, spacing)\n"
"    end\n"
"    return table.concat(rows, ass_escape('\\n')), row_count\n"
"end\n"
"\n"
"local function fuzzy_find(needle, haystacks)\n"
"    local result = require 'mp.fzy'.filter(needle, haystacks)\n"
"    if line ~= '' then -- Prevent table.sort() from reordering the items.\n"
"        table.sort(result, function (i, j)\n"
"            return i[3] > j[3]\n"
"        end)\n"
"    end\n"
"    for i, value in ipairs(result) do\n"
"        result[i] = value[1]\n"
"    end\n"
"    return result\n"
"end\n"
"\n"
"local function calculate_max_terminal_width()\n"
"    local max_width = mp.get_property_native('term-size/w', 80)\n"
"\n"
"    -- The longest module name is vo/gpu-next/libplacebo.\n"
"    if mp.get_property_native('msg-module') then\n"
"        max_width = max_width - 24\n"
"    end\n"
"\n"
"    if mp.get_property_native('msg-time') then\n"
"        max_width = max_width - 13\n"
"    end\n"
"\n"
"    return max_width\n"
"end\n"
"\n"
"local function populate_log_with_matches(max_width)\n"
"    if not selectable_items or selected_match == 0 then\n"
"        return\n"
"    end\n"
"\n"
"    log_buffers[id] = {}\n"
"    local log = log_buffers[id]\n"
"\n"
"    local max_log_lines = calculate_max_log_lines()\n"
"\n"
"    if selected_match < first_match_to_print then\n"
"        first_match_to_print = selected_match\n"
"    end\n"
"\n"
"    if first_match_to_print > 1 then\n"
"        -- Reserve the first line for \"n hidden items\".\n"
"        max_log_lines = max_log_lines - 1\n"
"    end\n"
"\n"
"    if selected_match > first_match_to_print + max_log_lines - 1 then\n"
"        -- Reserve the first line for \"n hidden items\" if it wasn't already.\n"
"        if first_match_to_print == 1 then\n"
"            max_log_lines = max_log_lines - 1\n"
"        end\n"
"\n"
"        first_match_to_print = selected_match - max_log_lines + 1\n"
"    end\n"
"\n"
"    local last_match_to_print  = math.min(first_match_to_print + max_log_lines - 1,\n"
"                                          #matches)\n"
"\n"
"    if last_match_to_print < #matches then\n"
"        -- Reserve the last line for \"n hidden items\".\n"
"        last_match_to_print = last_match_to_print - 1\n"
"\n"
"        -- After decrementing the last match to print, we need to check if the\n"
"        -- selected match is beyond the last match to print again, and shift\n"
"        -- both the first and last match to print when it is.\n"
"        if selected_match > last_match_to_print then\n"
"            if first_match_to_print == 1 then\n"
"                -- Reserve the first line for \"2 hidden items\".\n"
"                first_match_to_print = first_match_to_print + 1\n"
"            end\n"
"\n"
"            first_match_to_print = first_match_to_print + 1\n"
"            last_match_to_print = last_match_to_print + 1\n"
"        end\n"
"    end\n"
"\n"
"    -- When there is only 1 hidden item, print it in the previously reserved\n"
"    -- line instead of printing \"1 hidden items\".\n"
"    if first_match_to_print == 2 then\n"
"        first_match_to_print = 1\n"
"    end\n"
"    if last_match_to_print == #matches - 1 then\n"
"        last_match_to_print = #matches\n"
"    end\n"
"\n"
"    if first_match_to_print > 1 then\n"
"        log[1] = {\n"
"            text = '\342\206\221 (' .. (first_match_to_print - 1) .. ' hidden items)',\n"
"            style = styles.disabled,\n"
"            terminal_style = terminal_styles.disabled,\n"
"        }\n"
"    end\n"
"\n"
"    for i = first_match_to_print, last_match_to_print do\n"
"        log[#log + 1] = {\n"
"            text = (max_width and truncate_utf8(matches[i].text, max_width)\n"
"                    or matches[i].text),\n"
"            style = i == selected_match and styles.selected_suggestion or '',\n"
"            terminal_style = i == selected_match and terminal_styles.selected_suggestion or '',\n"
"        }\n"
"    end\n"
"\n"
"    if last_match_to_print < #matches then\n"
"        log[#log + 1] = {\n"
"            text = '\342\206\223 (' .. (#matches - last_match_to_print) .. ' hidden items)',\n"
"            style = styles.disabled,\n"
"            terminal_style = terminal_styles.disabled,\n"
"        }\n"
"    end\n"
"end\n"
"\n"
"local function print_to_terminal()\n"
"    -- Clear the log after closing the console.\n"
"    if not repl_active then\n"
"        mp.osd_message('')\n"
"        return\n"
"    end\n"
"\n"
"    populate_log_with_matches(calculate_max_terminal_width())\n"
"\n"
"    local log = ''\n"
"    for _, log_line in ipairs(log_buffers[id]) do\n"
"        log = log .. log_line.terminal_style .. log_line.text .. '\\027[0m\\n'\n"
"    end\n"
"\n"
"    local suggestions = ''\n"
"    for i, suggestion in ipairs(suggestion_buffer) do\n"
"        if i == selected_suggestion_index then\n"
"            suggestions = suggestions .. terminal_styles.selected_suggestion ..\n"
"                          suggestion .. '\\027[0m'\n"
"        else\n"
"            suggestions = suggestions .. suggestion\n"
"        end\n"
"        suggestions = suggestions .. (i < #suggestion_buffer and '\\t' or '\\n')\n"
"    end\n"
"\n"
"    local before_cur = line:sub(1, cursor - 1)\n"
"    local after_cur = line:sub(cursor)\n"
"    -- Ensure there is a character with inverted colors to print.\n"
"    if after_cur == '' then\n"
"        after_cur = ' '\n"
"    end\n"
"\n"
"    mp.osd_message(log .. suggestions .. prompt .. ' ' .. before_cur ..\n"
"                  '\\027[7m' .. after_cur:sub(1, 1) .. '\\027[0m' ..\n"
"                   after_cur:sub(2), 999)\n"
"end\n"
"\n"
"-- Render the REPL and console as an ASS OSD\n"
"local function update()\n"
"    pending_update = false\n"
"\n"
"    -- Unlike vo-configured, current-vo doesn't become falsy while switching VO,\n"
"    -- which would print the log to the OSD.\n"
"    if not mp.get_property('current-vo') or not mp.get_property_native('video-osd') then\n"
"        print_to_terminal()\n"
"        return\n"
"    end\n"
"\n"
"    -- Clear the OSD if the REPL is not active\n"
"    if not repl_active then\n"
"        mp.set_osd_ass(0, 0, '')\n"
"        return\n"
"    end\n"
"\n"
"    local screenx, screeny = mp.get_osd_size()\n"
"    local dpi_scale = mp.get_property_native('display-hidpi-scale', 1)\n"
"    screenx = screenx / dpi_scale\n"
"    screeny = screeny / dpi_scale\n"
"\n"
"    local bottom_left_margin = 6\n"
"\n"
"    local coordinate_top = math.floor(global_margins.t * screeny + 0.5)\n"
"    local clipping_coordinates = '0,' .. coordinate_top .. ',' ..\n"
"                                 screenx .. ',' .. screeny\n"
"    local ass = assdraw.ass_new()\n"
"    local has_shadow = mp.get_property('osd-border-style'):find('box\044') == nil\n"
"    local style = '{\\\\r' ..\n"
"                  '\\\\1a&H00&\\\\3a&H00&\\\\1c&Heeeeee&\\\\3c&H111111&' ..\n"
"                  (has_shadow and '\\\\4a&H99&\\\\4c&H000000&' or '') ..\n"
"                  '\\\\fn' .. opts.font .. '\\\\fs' .. opts.font_size ..\n"
"                  '\\\\bord' .. opts.border_size .. '\\\\xshad0\\\\yshad1\\\\fsp0' ..\n"
"                  (selectable_items and '\\\\q2' or '\\\\q1') ..\n"
"                  '\\\\clip(' .. clipping_coordinates .. ')}'\n"
"    -- Create the cursor glyph as an ASS drawing. ASS will draw the cursor\n"
"    -- inline with the surrounding text, but it sets the advance to the width\n"
"    -- of the drawing. So the cursor doesn't affect layout too much, make it as\n"
"    -- thin as possible and make it appear to be 1px wide by giving it 0.5px\n"
"    -- horizontal borders.\n"
"    local cheight = opts.font_size * 8\n"
"    local cglyph = '{\\\\rDefault' ..\n"
"                   (mp.get_property_native('focused') == false\n"
"                    and '\\\\alpha&HFF&' or '\\\\1a&H44&\\\\3a&H44&\\\\4a&H99&') ..\n"
"                   '\\\\1c&Heeeeee&\\\\3c&Heeeeee&\\\\4c&H000000&' ..\n"
"                   '\\\\xbord0.5\\\\ybord0\\\\xshad0\\\\yshad1\\\\p4\\\\pbo24}' ..\n"
"                   'm 0 0 l 1 0 l 1 ' .. cheight .. ' l 0 ' .. cheight ..\n"
"                   '{\\\\p0}'\n"
"    local before_cur = ass_escape(line:sub(1, cursor - 1))\n"
"    local after_cur = ass_escape(line:sub(cursor))\n"
"\n"
"    -- Render log messages as ASS.\n"
"    -- This will render at most screeny / font_size - 1 messages.\n"
"\n"
"    local lines_max = calculate_max_log_lines()\n"
"    -- Estimate how many characters fit in one line\n"
"    local width_max = math.floor((screenx - bottom_left_margin -\n"
"                                  mp.get_property_native('osd-margin-x') * 2 * screeny / 720) /\n"
"                                 opts.font_size * get_font_hw_ratio())\n"
"\n"
"    local suggestions, rows = format_table(suggestion_buffer, width_max, lines_max)\n"
"    local suggestion_ass = style .. styles.suggestion .. suggestions\n"
"\n"
"    populate_log_with_matches()\n"
"\n"
"    local log_ass = ''\n"
"    local log_buffer = log_buffers[id]\n"
"    local log_messages = #log_buffer\n"
"    local log_max_lines = math.max(0, lines_max - rows)\n"
"    if log_max_lines < log_messages then\n"
"        log_messages = log_max_lines\n"
"    end\n"
"    for i = #log_buffer - log_messages + 1, #log_buffer do\n"
"        log_ass = log_ass .. style .. log_buffer[i].style ..\n"
"                  ass_escape(log_buffer[i].text) .. '\\\\N'\n"
"    end\n"
"\n"
"    ass:new_event()\n"
"    ass:an(1)\n"
"    ass:pos(bottom_left_margin, screeny - bottom_left_margin - global_margins.b * screeny)\n"
"    ass:append(log_ass .. '\\\\N')\n"
"    if #suggestions > 0 then\n"
"        ass:append(suggestion_ass .. '\\\\N')\n"
"    end\n"
"    ass:append(style .. ass_escape(prompt) .. ' ' .. before_cur)\n"
"    ass:append(cglyph)\n"
"    ass:append(style .. after_cur)\n"
"\n"
"    -- Redraw the cursor with the REPL text invisible. This will make the\n"
"    -- cursor appear in front of the text.\n"
"    ass:new_event()\n"
"    ass:an(1)\n"
"    ass:pos(bottom_left_margin, screeny - bottom_left_margin - global_margins.b * screeny)\n"
"    ass:append(style .. '{\\\\alpha&HFF&}' .. ass_escape(prompt) .. ' ' .. before_cur)\n"
"    ass:append(cglyph)\n"
"    ass:append(style .. '{\\\\alpha&HFF&}' .. after_cur)\n"
"\n"
"    mp.set_osd_ass(screenx, screeny, ass.text)\n"
"end\n"
"\n"
"local update_timer = nil\n"
"update_timer = mp.add_periodic_timer(0.05, function()\n"
"    if pending_update then\n"
"        update()\n"
"    else\n"
"        update_timer:kill()\n"
"    end\n"
"end)\n"
"update_timer:kill()\n"
"\n"
"-- Add a line to the log buffer (which is limited to 100 lines)\n"
"local function log_add(text, style, terminal_style)\n"
"    local log_buffer = log_buffers[id]\n"
"    log_buffer[#log_buffer + 1] = {\n"
"        text = text,\n"
"        style = style or '',\n"
"        terminal_style = terminal_style or '',\n"
"    }\n"
"    if #log_buffer > 100 then\n"
"        table.remove(log_buffer, 1)\n"
"    end\n"
"\n"
"    if repl_active then\n"
"        if not update_timer:is_enabled() then\n"
"            update()\n"
"            update_timer:resume()\n"
"        else\n"
"            pending_update = true\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"-- Add a line to the history and deduplicate\n"
"local function history_add(text)\n"
"    if opts.history_dedup then\n"
"        -- More recent entries are more likely to be repeated\n"
"        for i = #history, 1, -1 do\n"
"            if history[i] == text then\n"
"                table.remove(history, i)\n"
"                break\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    history[#history + 1] = text\n"
"end\n"
"\n"
"local function handle_edit()\n"
"    if selectable_items then\n"
"        matches = {}\n"
"        selected_match = 1\n"
"\n"
"        for i, match in ipairs(fuzzy_find(line, selectable_items)) do\n"
"            matches[i] = { index = match, text = selectable_items[match] }\n"
"        end\n"
"    end\n"
"\n"
"    suggestion_buffer = {}\n"
"    update()\n"
"\n"
"    if input_caller then\n"
"        mp.commandv('script-message-to', input_caller, 'input-event', 'edited',\n"
"                    utils.format_json({line}))\n"
"    end\n"
"end\n"
"\n"
"-- Insert a character at the current cursor position (any_unicode)\n"
"local function handle_char_input(c)\n"
"    if insert_mode then\n"
"        line = line:sub(1, cursor - 1) .. c .. line:sub(next_utf8(line, cursor))\n"
"    else\n"
"        line = line:sub(1, cursor - 1) .. c .. line:sub(cursor)\n"
"    end\n"
"    cursor = cursor + #c\n"
"    handle_edit()\n"
"end\n"
"\n"
"-- Remove the character behind the cursor (Backspace)\n"
"local function handle_backspace()\n"
"    if cursor <= 1 then return end\n"
"    local prev = prev_utf8(line, cursor)\n"
"    line = line:sub(1, prev - 1) .. line:sub(cursor)\n"
"    cursor = prev\n"
"    handle_edit()\n"
"end\n"
"\n"
"-- Remove the character in front of the cursor (Del)\n"
"local function handle_del()\n"
"    if cursor > line:len() then return end\n"
"    line = line:sub(1, cursor - 1) .. line:sub(next_utf8(line, cursor))\n"
"    handle_edit()\n"
"end\n"
"\n"
"-- Toggle insert mode (Ins)\n"
"local function handle_ins()\n"
"    insert_mode = not insert_mode\n"
"end\n"
"\n"
"-- Move the cursor to the next character (Right)\n"
"local function next_char()\n"
"    cursor = next_utf8(line, cursor)\n"
"    suggestion_buffer = {}\n"
"    update()\n"
"end\n"
"\n"
"-- Move the cursor to the previous character (Left)\n"
"local function prev_char()\n"
"    cursor = prev_utf8(line, cursor)\n"
"    suggestion_buffer = {}\n"
"    update()\n"
"end\n"
"\n"
"-- Clear the current line (Ctrl+C)\n"
"local function clear()\n"
"    line = ''\n"
"    cursor = 1\n"
"    insert_mode = false\n"
"    history_pos = #history + 1\n"
"    handle_edit()\n"
"end\n"
"\n"
"-- Close the REPL if the current line is empty, otherwise delete the next\n"
"-- character (Ctrl+D)\n"
"local function maybe_exit()\n"
"    if line == '' then\n"
"        set_active(false)\n"
"    else\n"
"        handle_del()\n"
"    end\n"
"end\n"
"\n"
"local function help_command(param)\n"
"    local cmdlist = mp.get_property_native('command-list')\n"
"    table.sort(cmdlist, function(c1, c2)\n"
"        return c1.name < c2.name\n"
"    end)\n"
"    local output = ''\n"
"    if param == '' then\n"
"        output = 'Available commands:\\n'\n"
"        for _, cmd in ipairs(cmdlist) do\n"
"            output = output  .. '  ' .. cmd.name\n"
"        end\n"
"        output = output .. '\\n'\n"
"        output = output .. 'Use \"help command\" to show information about a command.\\n'\n"
"        output = output .. \"ESC or Ctrl+d exits the console.\\n\"\n"
"    else\n"
"        local cmd = nil\n"
"        for _, curcmd in ipairs(cmdlist) do\n"
"            if curcmd.name:find(param, 1, true) then\n"
"                cmd = curcmd\n"
"                if curcmd.name == param then\n"
"                    break -- exact match\n"
"                end\n"
"            end\n"
"        end\n"
"        if not cmd then\n"
"            log_add('No command matches \"' .. param .. '\"!', styles.error,\n"
"                    terminal_styles.error)\n"
"            return\n"
"        end\n"
"        output = output .. 'Command \"' .. cmd.name .. '\"\\n'\n"
"        for _, arg in ipairs(cmd.args) do\n"
"            output = output .. '    ' .. arg.name .. ' (' .. arg.type .. ')'\n"
"            if arg.optional then\n"
"                output = output .. ' (optional)'\n"
"            end\n"
"            output = output .. '\\n'\n"
"        end\n"
"        if cmd.vararg then\n"
"            output = output .. 'This command supports variable arguments.\\n'\n"
"        end\n"
"    end\n"
"    log_add(output:sub(1, -2))\n"
"end\n"
"\n"
"-- Run the current command and clear the line (Enter)\n"
"local function handle_enter()\n"
"    if searching_history then\n"
"        searching_history = false\n"
"        selectable_items = nil\n"
"        line = #matches > 0 and matches[selected_match].text or ''\n"
"        cursor = #line + 1\n"
"        log_buffers[id] = {}\n"
"        update()\n"
"        return\n"
"    end\n"
"\n"
"    if line == '' and input_caller == nil then\n"
"        return\n"
"    end\n"
"    if history[#history] ~= line and line ~= '' then\n"
"        history_add(line)\n"
"    end\n"
"\n"
"    if selectable_items then\n"
"        if #matches > 0 then\n"
"            mp.commandv('script-message-to', input_caller, 'input-event', 'submit',\n"
"                        utils.format_json({matches[selected_match].index}))\n"
"        end\n"
"        set_active(false)\n"
"    elseif input_caller then\n"
"        mp.commandv('script-message-to', input_caller, 'input-event', 'submit',\n"
"                    utils.format_json({line}))\n"
"    else\n"
"        -- match \"help [<text>]\", return <text> or \"\", strip all whitespace\n"
"        local help = line:match('^%s*help%s+(.-)%s*\044') or\n"
"                     (line:match('^%s*help\044') and '')\n"
"        if help then\n"
"            help_command(help)\n"
"        else\n"
"            mp.command(line)\n"
"        end\n"
"    end\n"
"\n"
"    clear()\n"
"end\n"
"\n"
"-- Go to the specified position in the command history\n"
"local function go_history(new_pos)\n"
"    local old_pos = history_pos\n"
"    history_pos = new_pos\n"
"\n"
"    -- Restrict the position to a legal value\n"
"    if history_pos > #history + 1 then\n"
"        history_pos = #history + 1\n"
"    elseif history_pos < 1 then\n"
"        history_pos = 1\n"
"    end\n"
"\n"
"    -- Do nothing if the history position didn't actually change\n"
"    if history_pos == old_pos then\n"
"        return\n"
"    end\n"
"\n"
"    -- If the user was editing a non-history line, save it as the last history\n"
"    -- entry. This makes it much less frustrating to accidentally hit Up/Down\n"
"    -- while editing a line.\n"
"    if old_pos == #history + 1 and line ~= '' and history[#history] ~= line then\n"
"        history_add(line)\n"
"    end\n"
"\n"
"    -- Now show the history line (or a blank line for #history + 1)\n"
"    if history_pos <= #history then\n"
"        line = history[history_pos]\n"
"    else\n"
"        line = ''\n"
"    end\n"
"    cursor = line:len() + 1\n"
"    insert_mode = false\n"
"    suggestion_buffer = {}\n"
"    update()\n"
"end\n"
"\n"
"-- Go to the specified relative position in the command history (Up, Down)\n"
"local function move_history(amount)\n"
"    if selectable_items then\n"
"        selected_match = selected_match + amount\n"
"        if selected_match > #matches then\n"
"            selected_match = 1\n"
"        elseif selected_match < 1 then\n"
"            selected_match = #matches\n"
"        end\n"
"        update()\n"
"        return\n"
"    end\n"
"\n"
"    go_history(history_pos + amount)\n"
"end\n"
"\n"
"-- Go to the first command in the command history (PgUp)\n"
"local function handle_pgup()\n"
"    if selectable_items then\n"
"        -- We don't know whether to count the \"n hidden items\" lines here; an\n"
"        -- offset of 2 is better with 1 extra line because it scrolls from the\n"
"        -- last to the first visible match, while with both extra lines that is\n"
"        -- done with +3. When there are no \"n hidden items\" lines selected_match\n"
"        -- becomes 1 with any offset >= 1.\n"
"        selected_match = math.max(selected_match - calculate_max_log_lines() + 2, 1)\n"
"        update()\n"
"        return\n"
"    end\n"
"\n"
"    go_history(1)\n"
"end\n"
"\n"
"-- Stop browsing history and start editing a blank line (PgDown)\n"
"local function handle_pgdown()\n"
"    if selectable_items then\n"
"        selected_match = math.min(selected_match + calculate_max_log_lines() - 2, #matches)\n"
"        update()\n"
"        return\n"
"    end\n"
"\n"
"    go_history(#history + 1)\n"
"end\n"
"\n"
"local function search_history()\n"
"    if selectable_items or #history == 0 then\n"
"        return\n"
"    end\n"
"\n"
"    searching_history = true\n"
"    selectable_items = {}\n"
"    matches = {}\n"
"    selected_match = 1\n"
"    first_match_to_print = 1\n"
"\n"
"    for i = 1, #history do\n"
"        selectable_items[i] = history[#history + 1 - i]\n"
"    end\n"
"\n"
"    for i, match in ipairs(fuzzy_find(line, selectable_items)) do\n"
"        matches[i] = { index = match, text = selectable_items[match] }\n"
"    end\n"
"\n"
"    update()\n"
"end\n"
"\n"
"local function page_up_or_prev_char()\n"
"    if selectable_items then\n"
"        handle_pgup()\n"
"    else\n"
"        prev_char()\n"
"    end\n"
"end\n"
"\n"
"local function page_down_or_next_char()\n"
"    if selectable_items then\n"
"        handle_pgdown()\n"
"    else\n"
"        next_char()\n"
"    end\n"
"end\n"
"\n"
"-- Move to the start of the current word, or if already at the start, the start\n"
"-- of the previous word. (Ctrl+Left)\n"
"local function prev_word()\n"
"    -- This is basically the same as next_word() but backwards, so reverse the\n"
"    -- string in order to do a \"backwards\" find. This wouldn't be as annoying\n"
"    -- to do if Lua didn't insist on 1-based indexing.\n"
"    cursor = line:len() - select(2, line:reverse():find('%s*[^%s]*', line:len() - cursor + 2)) + 1\n"
"    suggestion_buffer = {}\n"
"    update()\n"
"end\n"
"\n"
"-- Move to the end of the current word, or if already at the end, the end of\n"
"-- the next word. (Ctrl+Right)\n"
"local function next_word()\n"
"    cursor = select(2, line:find('%s*[^%s]*', cursor)) + 1\n"
"    suggestion_buffer = {}\n"
"    update()\n"
"end\n"
"\n"
"-- Move the cursor to the beginning of the line (HOME)\n"
"local function go_home()\n"
"    cursor = 1\n"
"    suggestion_buffer = {}\n"
"    update()\n"
"end\n"
"\n"
"-- Move the cursor to the end of the line (END)\n"
"local function go_end()\n"
"    cursor = line:len() + 1\n"
"    suggestion_buffer = {}\n"
"    update()\n"
"end\n"
"\n"
"-- Delete from the cursor to the beginning of the word (Ctrl+Backspace)\n"
"local function del_word()\n"
"    local before_cur = line:sub(1, cursor - 1)\n"
"    local after_cur = line:sub(cursor)\n"
"\n"
"    before_cur = before_cur:gsub('[^%s]+%s*\044', '', 1)\n"
"    line = before_cur .. after_cur\n"
"    cursor = before_cur:len() + 1\n"
"    handle_edit()\n"
"end\n"
"\n"
"-- Delete from the cursor to the end of the word (Ctrl+Del)\n"
"local function del_next_word()\n"
"    if cursor > line:len() then return end\n"
"\n"
"    local before_cur = line:sub(1, cursor - 1)\n"
"    local after_cur = line:sub(cursor)\n"
"\n"
"    after_cur = after_cur:gsub('^%s*[^%s]+', '', 1)\n"
"    line = before_cur .. after_cur\n"
"    handle_edit()\n"
"end\n"
"\n"
"-- Delete from the cursor to the end of the line (Ctrl+K)\n"
"local function del_to_eol()\n"
"    line = line:sub(1, cursor - 1)\n"
"    handle_edit()\n"
"end\n"
"\n"
"-- Delete from the cursor back to the start of the line (Ctrl+U)\n"
"local function del_to_start()\n"
"    line = line:sub(cursor)\n"
"    cursor = 1\n"
"    handle_edit()\n"
"end\n"
"\n"
"-- Empty the log buffer of all messages (Ctrl+L)\n"
"local function clear_log_buffer()\n"
"    log_buffers[id] = {}\n"
"    update()\n"
"end\n"
"\n"
"-- Returns a string of UTF-8 text from the clipboard (or the primary selection)\n"
"local function get_clipboard(clip)\n"
"    if platform == 'x11' then\n"
"        local res = utils.subprocess({\n"
"            args = { 'xclip', '-selection', clip and 'clipboard' or 'primary', '-out' },\n"
"            playback_only = false,\n"
"        })\n"
"        if not res.error then\n"
"            return res.stdout\n"
"        end\n"
"    elseif platform == 'wayland' then\n"
"        local res = utils.subprocess({\n"
"            args = { 'wl-paste', clip and '-n' or  '-np' },\n"
"            playback_only = false,\n"
"        })\n"
"        if not res.error then\n"
"            return res.stdout\n"
"        end\n"
"    elseif platform == 'windows' then\n"
"        local res = utils.subprocess({\n"
"            args = { 'powershell', '-NoProfile', '-Command', [[& {\n"
"                Trap {\n"
"                    Write-Error -ErrorRecord \044_\n"
"                    Exit 1\n"
"                }\n"
"\n"
"                \044clip = \"\"\n"
"                if (Get-Command \"Get-Clipboard\" -errorAction SilentlyContinue) {\n"
"                    \044clip = Get-Clipboard -Raw -Format Text -TextFormatType UnicodeText\n"
"                } else {\n"
"                    Add-Type -AssemblyName PresentationCore\n"
"                    \044clip = [Windows.Clipboard]::GetText()\n"
"                }\n"
"\n"
"                \044clip = \044clip -Replace \"\140r\",\"\"\n"
"                \044u8clip = [System.Text.Encoding]::UTF8.GetBytes(\044clip)\n"
"                [Console]::OpenStandardOutput().Write(\044u8clip, 0, \044u8clip.Length)\n"
"            }]] },\n"
"            playback_only = false,\n"
"        })\n"
"        if not res.error then\n"
"            return res.stdout\n"
"        end\n"
"    elseif platform == 'darwin' then\n"
"        local res = utils.subprocess({\n"
"            args = { 'pbpaste' },\n"
"            playback_only = false,\n"
"        })\n"
"        if not res.error then\n"
"            return res.stdout\n"
"        end\n"
"    end\n"
"    return ''\n"
"end\n"
"\n"
"-- Paste text from the window-system's clipboard. 'clip' determines whether the\n"
"-- clipboard or the primary selection buffer is used (on X11 and Wayland only.)\n"
"local function paste(clip)\n"
"    local text = get_clipboard(clip)\n"
"    local before_cur = line:sub(1, cursor - 1)\n"
"    local after_cur = line:sub(cursor)\n"
"    line = before_cur .. text .. after_cur\n"
"    cursor = cursor + text:len()\n"
"    handle_edit()\n"
"end\n"
"\n"
"local function text_input(info)\n"
"    if info.key_text and (info.event == \"press\" or info.event == \"down\"\n"
"                          or info.event == \"repeat\")\n"
"    then\n"
"        handle_char_input(info.key_text)\n"
"    end\n"
"end\n"
"\n"
"local function command_list()\n"
"    local commands = {}\n"
"    for i, command in ipairs(mp.get_property_native('command-list')) do\n"
"        commands[i] = command.name\n"
"    end\n"
"\n"
"    return commands\n"
"end\n"
"\n"
"local function property_list()\n"
"    local properties = mp.get_property_native('property-list')\n"
"\n"
"    for _, sub_property in pairs({'video', 'audio', 'sub', 'sub2'}) do\n"
"        properties[#properties + 1] = 'current-tracks/' .. sub_property\n"
"    end\n"
"\n"
"    for _, option in ipairs(mp.get_property_native('options')) do\n"
"        properties[#properties + 1] = 'options/' .. option\n"
"        properties[#properties + 1] = 'file-local-options/' .. option\n"
"        properties[#properties + 1] = 'option-info/' .. option\n"
"\n"
"        for _, sub_property in pairs({\n"
"            'name', 'type', 'set-from-commandline', 'set-locally',\n"
"            'expects-file', 'default-value', 'min', 'max', 'choices',\n"
"        }) do\n"
"            properties[#properties + 1] = 'option-info/' .. option .. '/' ..\n"
"                                          sub_property\n"
"        end\n"
"    end\n"
"\n"
"    return properties\n"
"end\n"
"\n"
"local function profile_list()\n"
"    local profiles = {}\n"
"\n"
"    for i, profile in ipairs(mp.get_property_native('profile-list')) do\n"
"        profiles[i] = profile.name\n"
"    end\n"
"\n"
"    return profiles\n"
"end\n"
"\n"
"local function list_option_list()\n"
"    local options = {}\n"
"\n"
"    -- Don't log errors for renamed and removed properties.\n"
"    -- (Just mp.enable_messages('fatal') still logs them to the terminal.)\n"
"    local msg_level_backup = mp.get_property('msg-level')\n"
"    mp.set_property('msg-level', msg_level_backup == '' and 'cplayer=no'\n"
"                                 or msg_level_backup .. ',cplayer=no')\n"
"\n"
"    for _, option in pairs(mp.get_property_native('options')) do\n"
"        if mp.get_property('option-info/' .. option .. '/type', ''):find(' list\044') then\n"
"            options[#options + 1] = option\n"
"        end\n"
"    end\n"
"\n"
"    mp.set_property('msg-level', msg_level_backup)\n"
"\n"
"    return options\n"
"end\n"
"\n"
"local function list_option_action_list(option)\n"
"    local type = mp.get_property('option-info/' .. option .. '/type')\n"
"\n"
"    if type == 'Key/value list' then\n"
"        return {'add', 'append', 'set', 'remove'}\n"
"    end\n"
"\n"
"    if type == 'String list' or type == 'Object settings list' then\n"
"        return {'add', 'append', 'clr', 'pre', 'set', 'remove', 'toggle'}\n"
"    end\n"
"end\n"
"\n"
"local function list_option_value_list(option)\n"
"    local values = mp.get_property_native(option)\n"
"\n"
"    if type(values) ~= 'table' then\n"
"        return\n"
"    end\n"
"\n"
"    if type(values[1]) ~= 'table' then\n"
"        return values\n"
"    end\n"
"\n"
"    for i, value in ipairs(values) do\n"
"        values[i] = value.label and '\100' .. value.label or value.name\n"
"    end\n"
"\n"
"    return values\n"
"end\n"
"\n"
"local function has_file_argument(candidate_command)\n"
"    for _, command in pairs(mp.get_property_native('command-list')) do\n"
"        if command.name == candidate_command then\n"
"            return command.args[1] and\n"
"                   (command.args[1].name == 'filename' or command.args[1].name == 'url')\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function file_list(directory)\n"
"    if directory == '' then\n"
"        directory = '.'\n"
"    end\n"
"\n"
"    local files = utils.readdir(directory, 'files') or {}\n"
"\n"
"    for _, dir in pairs(utils.readdir(directory, 'dirs') or {}) do\n"
"        files[#files + 1] = dir .. path_separator\n"
"    end\n"
"\n"
"    return files\n"
"end\n"
"\n"
"local function handle_file_completion(before_cur, path_pos)\n"
"    local directory, last_component_pos =\n"
"        before_cur:sub(path_pos):match('(.-)()[^' .. path_separator ..']*\044')\n"
"    completion_pos = path_pos + last_component_pos - 1\n"
"\n"
"    if directory:find('^~' .. path_separator) then\n"
"        local home = mp.command_native({'expand-path', '~/'})\n"
"        before_cur = before_cur:sub(1, completion_pos - #directory - 1) ..\n"
"                     home ..\n"
"                     before_cur:sub(completion_pos - #directory + 1)\n"
"        directory = home .. directory:sub(2)\n"
"        completion_pos = completion_pos + #home - 1\n"
"    end\n"
"\n"
"    -- Don't use completion_append for file completion to not add quotes after\n"
"    -- directories whose entries you may want to complete afterwards.\n"
"    completion_append = ''\n"
"\n"
"    return file_list(directory), before_cur\n"
"end\n"
"\n"
"local function handle_choice_completion(option, before_cur, path_pos)\n"
"    local info = mp.get_property_native('option-info/' .. option, {})\n"
"\n"
"    if info.type == 'Flag' then\n"
"        return { 'no', 'yes' }, before_cur\n"
"    end\n"
"\n"
"    if info['expects-file'] then\n"
"        return handle_file_completion(before_cur, path_pos)\n"
"    end\n"
"\n"
"    -- Fix completing the empty value for --dscale and --cscale.\n"
"    if info.choices and info.choices[1] == '' and completion_append == '' then\n"
"        info.choices[1] = '\"\"'\n"
"    end\n"
"\n"
"    return info.choices, before_cur\n"
"end\n"
"\n"
"local function common_prefix_length(s1, s2)\n"
"    local common_count = 0\n"
"    for i = 1, #s1 do\n"
"        if s1:byte(i) ~= s2:byte(i) then\n"
"            break\n"
"        end\n"
"        common_count = common_count + 1\n"
"    end\n"
"    return common_count\n"
"end\n"
"\n"
"local function max_overlap_length(s1, s2)\n"
"    for s1_offset = 0, #s1 - 1 do\n"
"        local match = true\n"
"        for i = 1, #s1 - s1_offset do\n"
"            if s1:byte(s1_offset + i) ~= s2:byte(i) then\n"
"                match = false\n"
"                break\n"
"            end\n"
"        end\n"
"        if match then\n"
"            return #s1 - s1_offset\n"
"        end\n"
"    end\n"
"    return 0\n"
"end\n"
"\n"
"-- If str starts with the first or last characters of prefix, strip them.\n"
"local function strip_common_characters(str, prefix)\n"
"    return str:sub(1 + math.max(\n"
"    common_prefix_length(prefix, str),\n"
"    max_overlap_length(prefix, str)))\n"
"end\n"
"\n"
"-- Find the longest common case-sensitive prefix of the entries in \"list\".\n"
"local function find_common_prefix(list)\n"
"    local prefix = list[1]\n"
"\n"
"    for i = 2, #list do\n"
"        prefix = prefix:sub(1, common_prefix_length(prefix, list[i]))\n"
"    end\n"
"\n"
"    return prefix\n"
"end\n"
"\n"
"-- Return the entries of \"list\" beginning with \"part\" and the longest common\n"
"-- prefix of the matches.\n"
"local function complete_match(part, list)\n"
"    local completions = {}\n"
"\n"
"    for _, candidate in pairs(list) do\n"
"        if candidate:sub(1, part:len()) == part then\n"
"            completions[#completions + 1] = candidate\n"
"        end\n"
"    end\n"
"\n"
"    local prefix = find_common_prefix(completions)\n"
"\n"
"    if opts.case_sensitive then\n"
"        return completions, prefix or part\n"
"    end\n"
"\n"
"    completions = {}\n"
"    local lower_case_completions = {}\n"
"    local lower_case_part = part:lower()\n"
"\n"
"    for _, candidate in pairs(list) do\n"
"        if candidate:sub(1, part:len()):lower() == lower_case_part then\n"
"            completions[#completions + 1] = candidate\n"
"            lower_case_completions[#lower_case_completions + 1] = candidate:lower()\n"
"        end\n"
"    end\n"
"\n"
"    local lower_case_prefix = find_common_prefix(lower_case_completions)\n"
"\n"
"    -- Behave like GNU readline with completion-ignore-case On.\n"
"    -- part = 'fooBA', completions = {'foobarbaz', 'fooBARqux'} =>\n"
"    -- prefix = 'fooBARqux', lower_case_prefix = 'foobar', return 'fooBAR'\n"
"    if prefix then\n"
"        return completions, prefix:sub(1, lower_case_prefix:len())\n"
"    end\n"
"\n"
"    -- part = 'fooba', completions = {'fooBARbaz', 'fooBarqux'} =>\n"
"    -- prefix = nil, lower_case_prefix ='foobar', return 'fooBAR'\n"
"    if lower_case_prefix then\n"
"        return completions, completions[1]:sub(1, lower_case_prefix:len())\n"
"    end\n"
"\n"
"    return {}, part\n"
"end\n"
"\n"
"local function cycle_through_suggestions(backwards)\n"
"    selected_suggestion_index = selected_suggestion_index + (backwards and -1 or 1)\n"
"\n"
"    if selected_suggestion_index > #suggestion_buffer then\n"
"        selected_suggestion_index = 1\n"
"    elseif selected_suggestion_index < 1 then\n"
"        selected_suggestion_index = #suggestion_buffer\n"
"    end\n"
"\n"
"    local before_cur = line:sub(1, completion_pos - 1) ..\n"
"                       suggestion_buffer[selected_suggestion_index] .. completion_append\n"
"    line = before_cur .. strip_common_characters(line:sub(cursor), completion_append)\n"
"    cursor = before_cur:len() + 1\n"
"    update()\n"
"end\n"
"\n"
"-- Complete the option or property at the cursor (TAB)\n"
"local function complete(backwards)\n"
"    if #suggestion_buffer > 0 then\n"
"        cycle_through_suggestions(backwards)\n"
"        return\n"
"    end\n"
"\n"
"    if input_caller then\n"
"        completion_old_line = line\n"
"        completion_old_cursor = cursor\n"
"        mp.commandv('script-message-to', input_caller, 'input-event',\n"
"                    'complete', utils.format_json({line:sub(1, cursor - 1)}))\n"
"        return\n"
"    end\n"
"\n"
"    local before_cur = line:sub(1, cursor - 1)\n"
"    local after_cur = line:sub(cursor)\n"
"    local tokens = {}\n"
"    local first_useful_token_index = 1\n"
"    local completions\n"
"\n"
"    local begin_new_token = true\n"
"    local last_quote\n"
"    for pos, char in before_cur:gmatch('()(.)') do\n"
"        if char:find('[%s;]') and not last_quote then\n"
"            begin_new_token = true\n"
"            if char == ';' then\n"
"                first_useful_token_index = #tokens + 1\n"
"            end\n"
"        elseif begin_new_token then\n"
"            tokens[#tokens + 1] = { text = char, pos = pos }\n"
"            last_quote = char:match('[\"\\']')\n"
"            begin_new_token = false\n"
"        else\n"
"            tokens[#tokens].text = tokens[#tokens].text .. char\n"
"            if char == last_quote then\n"
"                last_quote = nil\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    completion_append = last_quote or ''\n"
"\n"
"    -- Strip quotes from tokens.\n"
"    for _, token in pairs(tokens) do\n"
"        if token.text:find('^\"') then\n"
"            token.text = token.text:sub(2):gsub('\"\044', '')\n"
"            token.pos = token.pos + 1\n"
"        elseif token.text:find(\"^'\") then\n"
"            token.text = token.text:sub(2):gsub(\"'\044\", '')\n"
"            token.pos = token.pos + 1\n"
"        end\n"
"    end\n"
"\n"
"    -- Skip command prefixes because it is not worth lumping them together with\n"
"    -- command completions when they are useless for interactive usage.\n"
"    local command_prefixes = {\n"
"        ['osd-auto'] = true, ['no-osd'] = true, ['osd-bar'] = true,\n"
"        ['osd-msg'] = true, ['osd-msg-bar'] = true, ['raw'] = true,\n"
"        ['expand-properties'] = true, ['repeatable'] = true,\n"
"        ['nonrepeatable'] = true, ['async'] = true, ['sync'] = true\n"
"    }\n"
"\n"
"    while tokens[first_useful_token_index] and\n"
"          command_prefixes[tokens[first_useful_token_index].text] do\n"
"        first_useful_token_index = first_useful_token_index + 1\n"
"    end\n"
"\n"
"    -- Add an empty token if the cursor is after whitespace to simplify\n"
"    -- comparisons.\n"
"    if before_cur == '' or before_cur:find('[%s;]\044') then\n"
"        tokens[#tokens + 1] = { text = \"\", pos = cursor }\n"
"    end\n"
"\n"
"    local add_actions = {\n"
"        ['add'] = true, ['append'] = true, ['pre'] = true, ['set'] = true\n"
"    }\n"
"\n"
"    local first_useful_token = tokens[first_useful_token_index]\n"
"\n"
"    completion_pos = before_cur:match('\044{[=>]\077()[%w_/-]*\044')\n"
"    if completion_pos then\n"
"        completions = property_list()\n"
"        completion_append = '} '\n"
"    elseif #tokens == first_useful_token_index then\n"
"        completions = command_list()\n"
"        completions[#completions + 1] = 'help'\n"
"        completion_pos = first_useful_token.pos\n"
"        completion_append = completion_append .. ' '\n"
"    elseif #tokens == first_useful_token_index + 1 then\n"
"        if first_useful_token.text == 'set' or\n"
"           first_useful_token.text == 'add' or\n"
"           first_useful_token.text == 'cycle' or\n"
"           first_useful_token.text == 'cycle-values' or\n"
"           first_useful_token.text == 'multiply' then\n"
"            completions = property_list()\n"
"            completion_pos = tokens[first_useful_token_index + 1].pos\n"
"            completion_append = completion_append .. ' '\n"
"        elseif first_useful_token.text == 'help' then\n"
"            completions = command_list()\n"
"            completion_pos = tokens[first_useful_token_index + 1].pos\n"
"        elseif first_useful_token.text == 'apply-profile' then\n"
"            completions = profile_list()\n"
"            completion_pos = tokens[first_useful_token_index + 1].pos\n"
"        elseif first_useful_token.text == 'change-list' then\n"
"            completions = list_option_list()\n"
"            completion_pos = tokens[first_useful_token_index + 1].pos\n"
"            completion_append = completion_append .. ' '\n"
"        elseif first_useful_token.text == 'vf' or\n"
"               first_useful_token.text == 'af' then\n"
"            completions = list_option_action_list(first_useful_token.text)\n"
"            completion_pos = tokens[first_useful_token_index + 1].pos\n"
"            completion_append = completion_append .. ' '\n"
"        elseif has_file_argument(first_useful_token.text) then\n"
"            completions, before_cur =\n"
"                handle_file_completion(before_cur, tokens[first_useful_token_index + 1].pos)\n"
"        end\n"
"    elseif first_useful_token.text == 'cycle-values' then\n"
"        completion_pos = tokens[#tokens].pos\n"
"        completion_append = completion_append .. ' '\n"
"        completions, before_cur =\n"
"            handle_choice_completion(tokens[first_useful_token_index + 1].text,\n"
"                                     before_cur, tokens[#tokens].pos)\n"
"    elseif #tokens == first_useful_token_index + 2 then\n"
"        if first_useful_token.text == 'set' then\n"
"            completion_pos = tokens[#tokens].pos\n"
"            completions, before_cur =\n"
"                handle_choice_completion(tokens[first_useful_token_index + 1].text,\n"
"                                         before_cur,\n"
"                                         tokens[first_useful_token_index + 2].pos)\n"
"        elseif first_useful_token.text == 'change-list' then\n"
"            completions = list_option_action_list(tokens[first_useful_token_index + 1].text)\n"
"            completion_pos = tokens[first_useful_token_index + 2].pos\n"
"            completion_append = completion_append .. ' '\n"
"        elseif first_useful_token.text == 'vf' or\n"
"               first_useful_token.text == 'af' then\n"
"            if add_actions[tokens[first_useful_token_index + 1].text] then\n"
"                completion_pos = tokens[#tokens].pos\n"
"                completions, before_cur =\n"
"                    handle_choice_completion(first_useful_token.text,\n"
"                                             before_cur, tokens[#tokens].pos)\n"
"            elseif tokens[first_useful_token_index + 1].text == 'remove' then\n"
"                completions = list_option_value_list(first_useful_token.text)\n"
"                completion_pos = tokens[#tokens].pos\n"
"            end\n"
"        end\n"
"    elseif #tokens == first_useful_token_index + 3 then\n"
"        if first_useful_token.text == 'change-list' then\n"
"            if add_actions[tokens[first_useful_token_index + 2].text] then\n"
"                completion_pos = tokens[#tokens].pos\n"
"                completions, before_cur =\n"
"                    handle_choice_completion(tokens[first_useful_token_index + 1].text,\n"
"                                             before_cur, tokens[#tokens].pos)\n"
"            elseif tokens[first_useful_token_index + 2].text == 'remove' then\n"
"                completion_pos = tokens[#tokens].pos\n"
"                completions = list_option_value_list(tokens[first_useful_token_index + 1].text)\n"
"            end\n"
"        elseif first_useful_token.text == 'dump-cache' then\n"
"            completions, before_cur =\n"
"                handle_file_completion(before_cur,\n"
"                                       tokens[first_useful_token_index + 3].pos)\n"
"        end\n"
"    end\n"
"\n"
"    if completions == nil then\n"
"        return\n"
"    end\n"
"\n"
"    local prefix\n"
"    completions, prefix =\n"
"        complete_match(before_cur:sub(completion_pos), completions)\n"
"\n"
"    if #completions == 1 then\n"
"        prefix = prefix .. completion_append\n"
"        after_cur = strip_common_characters(after_cur, completion_append)\n"
"    else\n"
"        table.sort(completions)\n"
"        suggestion_buffer = completions\n"
"        selected_suggestion_index = 0\n"
"    end\n"
"\n"
"    before_cur = before_cur:sub(1, completion_pos - 1) .. prefix\n"
"    cursor = before_cur:len() + 1\n"
"    line = before_cur .. after_cur\n"
"    update()\n"
"end\n"
"\n"
"-- List of input bindings. This is a weird mashup between common GUI text-input\n"
"-- bindings and readline bindings.\n"
"local function get_bindings()\n"
"    local bindings = {\n"
"        { 'esc',         function() set_active(false) end       },\n"
"        { 'ctrl+[',      function() set_active(false) end       },\n"
"        { 'enter',       handle_enter                           },\n"
"        { 'kp_enter',    handle_enter                           },\n"
"        { 'shift+enter', function() handle_char_input('\\n') end },\n"
"        { 'ctrl+j',      handle_enter                           },\n"
"        { 'ctrl+m',      handle_enter                           },\n"
"        { 'bs',          handle_backspace                       },\n"
"        { 'shift+bs',    handle_backspace                       },\n"
"        { 'ctrl+h',      handle_backspace                       },\n"
"        { 'del',         handle_del                             },\n"
"        { 'shift+del',   handle_del                             },\n"
"        { 'ins',         handle_ins                             },\n"
"        { 'shift+ins',   function() paste(false) end            },\n"
"        { 'mbtn_mid',    function() paste(false) end            },\n"
"        { 'left',        function() prev_char() end             },\n"
"        { 'ctrl+b',      function() page_up_or_prev_char() end  },\n"
"        { 'right',       function() next_char() end             },\n"
"        { 'ctrl+f',      function() page_down_or_next_char() end},\n"
"        { 'up',          function() move_history(-1) end        },\n"
"        { 'ctrl+p',      function() move_history(-1) end        },\n"
"        { 'wheel_up',    function() move_history(-1) end        },\n"
"        { 'down',        function() move_history(1) end         },\n"
"        { 'ctrl+n',      function() move_history(1) end         },\n"
"        { 'wheel_down',  function() move_history(1) end         },\n"
"        { 'wheel_left',  function() end                         },\n"
"        { 'wheel_right', function() end                         },\n"
"        { 'ctrl+left',   prev_word                              },\n"
"        { 'alt+b',       prev_word                              },\n"
"        { 'ctrl+right',  next_word                              },\n"
"        { 'alt+f',       next_word                              },\n"
"        { 'tab',         complete                               },\n"
"        { 'ctrl+i',      complete                               },\n"
"        { 'shift+tab',   function() complete(true) end          },\n"
"        { 'ctrl+a',      go_home                                },\n"
"        { 'home',        go_home                                },\n"
"        { 'ctrl+e',      go_end                                 },\n"
"        { 'end',         go_end                                 },\n"
"        { 'pgup',        handle_pgup                            },\n"
"        { 'pgdwn',       handle_pgdown                          },\n"
"        { 'ctrl+r',      search_history                         },\n"
"        { 'ctrl+c',      clear                                  },\n"
"        { 'ctrl+d',      maybe_exit                             },\n"
"        { 'ctrl+k',      del_to_eol                             },\n"
"        { 'ctrl+l',      clear_log_buffer                       },\n"
"        { 'ctrl+u',      del_to_start                           },\n"
"        { 'ctrl+v',      function() paste(true) end             },\n"
"        { 'meta+v',      function() paste(true) end             },\n"
"        { 'ctrl+bs',     del_word                               },\n"
"        { 'ctrl+w',      del_word                               },\n"
"        { 'ctrl+del',    del_next_word                          },\n"
"        { 'alt+d',       del_next_word                          },\n"
"        { 'kp_dec',      function() handle_char_input('.') end  },\n"
"        { 'kp_add',      function() handle_char_input('+') end  },\n"
"        { 'kp_subtract', function() handle_char_input('-') end  },\n"
"        { 'kp_multiply', function() handle_char_input('*') end  },\n"
"        { 'kp_divide',   function() handle_char_input('/') end  },\n"
"    }\n"
"\n"
"    for i = 0, 9 do\n"
"        bindings[#bindings + 1] =\n"
"            {'kp' .. i, function() handle_char_input('' .. i) end}\n"
"    end\n"
"\n"
"    return bindings\n"
"end\n"
"\n"
"local function define_key_bindings()\n"
"    if #key_bindings > 0 then\n"
"        return\n"
"    end\n"
"    for _, bind in ipairs(get_bindings()) do\n"
"        -- Generate arbitrary name for removing the bindings later.\n"
"        local name = \"_console_\" .. (#key_bindings + 1)\n"
"        key_bindings[#key_bindings + 1] = name\n"
"        mp.add_forced_key_binding(bind[1], name, bind[2], {repeatable = true})\n"
"    end\n"
"    mp.add_forced_key_binding(\"any_unicode\", \"_console_text\", text_input,\n"
"        {repeatable = true, complex = true})\n"
"    key_bindings[#key_bindings + 1] = \"_console_text\"\n"
"end\n"
"\n"
"local function undefine_key_bindings()\n"
"    for _, name in ipairs(key_bindings) do\n"
"        mp.remove_key_binding(name)\n"
"    end\n"
"    key_bindings = {}\n"
"end\n"
"\n"
"-- Set the REPL visibility (\"enable\", Esc)\n"
"set_active = function (active)\n"
"    if active == repl_active then return end\n"
"    if active then\n"
"        repl_active = true\n"
"        insert_mode = false\n"
"        define_key_bindings()\n"
"\n"
"        if not input_caller then\n"
"            prompt = default_prompt\n"
"            id = default_id\n"
"            history = histories[id]\n"
"            history_pos = #history + 1\n"
"            mp.enable_messages('terminal-default')\n"
"        end\n"
"    elseif searching_history then\n"
"        searching_history = false\n"
"        line = ''\n"
"        cursor = 1\n"
"        selectable_items = nil\n"
"        log_buffers[id] = {}\n"
"    else\n"
"        repl_active = false\n"
"        suggestion_buffer = {}\n"
"        undefine_key_bindings()\n"
"        mp.enable_messages('silent:terminal-default')\n"
"\n"
"        if input_caller then\n"
"            mp.commandv('script-message-to', input_caller, 'input-event',\n"
"                        'closed', utils.format_json({line, cursor}))\n"
"            input_caller = nil\n"
"            line = ''\n"
"            cursor = 1\n"
"            selectable_items = nil\n"
"        end\n"
"        collectgarbage()\n"
"    end\n"
"    update()\n"
"end\n"
"\n"
"-- Show the repl if hidden and replace its contents with 'text'\n"
"-- (script-message-to repl type)\n"
"local function show_and_type(text, cursor_pos)\n"
"    text = text or ''\n"
"    cursor_pos = tonumber(cursor_pos)\n"
"\n"
"    -- Save the line currently being edited, just in case\n"
"    if line ~= text and line ~= '' and history[#history] ~= line then\n"
"        history_add(line)\n"
"    end\n"
"\n"
"    line = text\n"
"    if cursor_pos ~= nil and cursor_pos >= 1\n"
"        and cursor_pos <= line:len() + 1 then\n"
"        cursor = math.floor(cursor_pos)\n"
"    else\n"
"        cursor = line:len() + 1\n"
"    end\n"
"    history_pos = #history + 1\n"
"    insert_mode = false\n"
"    if repl_active then\n"
"        update()\n"
"    else\n"
"        set_active(true)\n"
"    end\n"
"end\n"
"\n"
"-- Add a global binding for enabling the REPL. While it's enabled, its bindings\n"
"-- will take over and it can be closed with ESC.\n"
"mp.add_key_binding(nil, 'enable', function()\n"
"    set_active(true)\n"
"end)\n"
"\n"
"mp.register_script_message('disable', function()\n"
"    set_active(false)\n"
"end)\n"
"\n"
"-- Add a script-message to show the REPL and fill it with the provided text\n"
"mp.register_script_message('type', function(text, cursor_pos)\n"
"    show_and_type(text, cursor_pos)\n"
"end)\n"
"\n"
"mp.register_script_message('get-input', function (script_name, args)\n"
"    if repl_active then\n"
"        return\n"
"    end\n"
"\n"
"    input_caller = script_name\n"
"    args = utils.parse_json(args)\n"
"    prompt = args.prompt or default_prompt\n"
"    line = args.default_text or ''\n"
"    cursor = args.cursor_position or line:len() + 1\n"
"    id = args.id or script_name .. prompt\n"
"    if histories[id] == nil then\n"
"        histories[id] = {}\n"
"        log_buffers[id] = {}\n"
"    end\n"
"    history = histories[id]\n"
"    history_pos = #history + 1\n"
"\n"
"    selectable_items = args.items\n"
"    if selectable_items then\n"
"        matches = {}\n"
"        selected_match = args.default_item or 1\n"
"        first_match_to_print = 1\n"
"        for i, item in ipairs(selectable_items) do\n"
"            matches[i] = { index = i, text = item }\n"
"        end\n"
"    end\n"
"\n"
"    set_active(true)\n"
"    mp.commandv('script-message-to', input_caller, 'input-event', 'opened')\n"
"end)\n"
"\n"
"mp.register_script_message('log', function (message)\n"
"    -- input.get's edited handler is invoked after submit, so avoid modifying\n"
"    -- the default log.\n"
"    if input_caller == nil then\n"
"        return\n"
"    end\n"
"\n"
"    message = utils.parse_json(message)\n"
"\n"
"    log_add(message.text,\n"
"            message.error and styles.error or message.style,\n"
"            message.error and terminal_styles.error or message.terminal_style)\n"
"end)\n"
"\n"
"mp.register_script_message('set-log', function (log)\n"
"    if input_caller == nil then\n"
"        return\n"
"    end\n"
"\n"
"    log = utils.parse_json(log)\n"
"    log_buffers[id] = {}\n"
"\n"
"    for i = 1, #log do\n"
"        if type(log[i]) == 'table' then\n"
"            log[i].text = log[i].text\n"
"            log[i].style = log[i].style or ''\n"
"            log[i].terminal_style = log[i].terminal_style or ''\n"
"            log_buffers[id][i] = log[i]\n"
"        else\n"
"            log_buffers[id][i] = {\n"
"                text = log[i],\n"
"                style = '',\n"
"                terminal_style = '',\n"
"            }\n"
"        end\n"
"    end\n"
"\n"
"    update()\n"
"end)\n"
"\n"
"mp.register_script_message('complete', function(list, start_pos)\n"
"    if line ~= completion_old_line or cursor ~= completion_old_cursor then\n"
"        return\n"
"    end\n"
"\n"
"    local completions, prefix = complete_match(line:sub(start_pos, cursor),\n"
"                                               utils.parse_json(list))\n"
"    local before_cur = line:sub(1, start_pos - 1) .. prefix\n"
"    local after_cur = line:sub(cursor)\n"
"    cursor = before_cur:len() + 1\n"
"    line = before_cur .. after_cur\n"
"\n"
"    if #completions > 1 then\n"
"        suggestion_buffer = completions\n"
"        selected_suggestion_index = 0\n"
"        completion_pos = start_pos\n"
"        completion_append = ''\n"
"    end\n"
"\n"
"    update()\n"
"end)\n"
"\n"
"-- Redraw the REPL when the OSD size changes. This is needed because the\n"
"-- PlayRes of the OSD will need to be adjusted.\n"
"mp.observe_property('osd-width', 'native', update)\n"
"mp.observe_property('osd-height', 'native', update)\n"
"mp.observe_property('display-hidpi-scale', 'native', update)\n"
"mp.observe_property('focused', 'native', update)\n"
"\n"
"mp.observe_property(\"user-data/osc/margins\", \"native\", function(_, val)\n"
"    if type(val) == \"table\" and type(val.t) == \"number\" and type(val.b) == \"number\" then\n"
"        global_margins = val\n"
"    else\n"
"        global_margins = { t = 0, b = 0 }\n"
"    end\n"
"    update()\n"
"end)\n"
"\n"
"-- Enable log messages. In silent mode, mpv will queue log messages in a buffer\n"
"-- until enable_messages is called again without the silent: prefix.\n"
"mp.enable_messages('silent:terminal-default')\n"
"\n"
"mp.register_event('log-message', function(e)\n"
"    -- Ignore log messages from the OSD because of paranoia, since writing them\n"
"    -- to the OSD could generate more messages in an infinite loop.\n"
"    if e.prefix:sub(1, 3) == 'osd' then return end\n"
"\n"
"    -- Ignore messages output by this script.\n"
"    if e.prefix == mp.get_script_name() then return end\n"
"\n"
"    -- Ignore buffer overflow warning messages. Overflowed log messages would\n"
"    -- have been offscreen anyway.\n"
"    if e.prefix == 'overflow' then return end\n"
"\n"
"    -- Filter out trace-level log messages, even if the terminal-default log\n"
"    -- level includes them. These aren't too useful for an on-screen display\n"
"    -- without scrollback and they include messages that are generated from the\n"
"    -- OSD display itself.\n"
"    if e.level == 'trace' then return end\n"
"\n"
"    -- Use color for debug/v/warn/error/fatal messages.\n"
"    log_add('[' .. e.prefix .. '] ' .. e.text:sub(1, -2), styles[e.level],\n"
"            terminal_styles[e.level])\n"
"end)\n"
"\n"
"collectgarbage()\n"
